import abc
from functools import singledispatch  # type: ignore

from amino import Path, __, Just, List, L, Maybe, _, Map

from ribosome.record import list_field, field, maybe_field, bool_field, dfield
from ribosome.util.callback import parse_callback_spec

from myo.record import Record, Named
from myo.util import ident_field, Ident
from myo.logging import Logging
from myo.util.ident import Key


default_signals = List('int', 'term', 'kill')


class Line(Logging, metaclass=abc.ABCMeta):

    def __init__(self, line) -> None:
        self.line = line

    @abc.abstractmethod
    def resolve(self, vim, args: List) -> Maybe[str]:
        ...


class StrictLine(Line):

    def resolve(self, vim, *args):
        return Just(self.line)


class EvalLine(Line):

    def resolve(self, vim, *args):
        return parse_callback_spec(self.line) // __(vim, *args)


class Command(Named):
    log_path = maybe_field(Path, factory=Path)
    parser = maybe_field(str)
    transient = bool_field()
    langs = list_field()
    kill = bool_field()
    signals = list_field(str, initial=default_signals)
    eval = bool_field()
    args = list_field()

    @property
    def _str_extra(self):
        return super()._str_extra.cat_m(self.langs.empty.no.m(self.langs))

    @property
    def _str_extra_named(self):
        return Map(eval=self.eval)

    @property
    def desc(self):
        info = self._info.mk_string(' ∘ ')
        tail = ': {}'.format(info) if info else ''
        return '{} {}{}'.format(self._type_desc, self.name, tail)

    @property
    def _info_fields(self):
        return List('langs', 'parser', 'log_path')

    @property
    def _info(self) -> List[str]:
        return self._info_fields // self._field_info

    def _field_info(self, name: str):
        v = getattr(self, name)
        fmt = '⎡{} -> {}⎤'.format
        if isinstance(v, Maybe):
            return v / L(fmt)(name, _)
        elif isinstance(v, List):
            return v.empty.no.maybe(fmt(name, v.mk_string(',')))
        else:
            return Just(fmt(name, v))

    @property
    def _type_desc(self):
        return 'C'

    @property
    def main_id(self) -> Ident:
        return self.uuid

    @property
    def main_name(self):
        return self.name

    @property
    def main_key(self):
        return Key(uuid=self.main_id, name=self.main_name)

    @property
    def effective(self):
        return self


class VimCommand(Command):

    @property
    def _type_desc(self):
        return 'VC'


class ShellCommand(Command):
    line = field(str)
    shell = ident_field()

    @property
    def _type_desc(self):
        return 'SC'

    @property
    def _info_fields(self):
        return super()._info_fields + List('shell')

    @property
    def effective_line(self):
        tpe = EvalLine if self.eval else StrictLine
        return tpe(self.line)


class Shell(ShellCommand):
    target = ident_field()

    @property
    def _type_desc(self):
        return 'S'


class TransientCommand(Command):
    command = field(Command)
    line = field(str)

    def __new__(cls, prefix='transient', transient=None, command=None,
                name=None, **kw):
        name = name or '{}_{}_{}'.format(prefix, command.name,
                                         List.random_string(5))
        return super().__new__(cls, name=name, transient=True, command=command,
                               **kw)

    @property
    def _str_extra(self):
        return super()._str_extra.cat(self.command)

    @property
    def effective(self):
        return self.command.set(name=self.name, line=self.line)

    @property
    def main_id(self):
        return self.command.main_id

    @property
    def main_name(self):
        return self.command.name


@singledispatch
def _history_cmp(cmd):
    return cmd.name


@_history_cmp.register(TransientCommand)
def _history_cmp_transient(cmd):
    return '{}|{}'.format(cmd.command.name, cmd.line)


class Commands(Record):
    max_history = dfield(20)

    no_such_command_error = 'no such command: {}'
    no_such_shell_error = 'so such shell: {}'
    no_latest_command_error = 'history is empty'

    commands = list_field()
    history = list_field()
    transient_cmd = maybe_field(Command)

    def __add__(self, command: Command):
        return self.append1.commands(command)

    def __getitem__(self, ident: Ident):
        return (
            self.commands.find(__.has_ident(ident))
            .o(lambda: self.history.find(__.has_ident(ident)))
            .to_either(self.no_such_command_error.format(ident))
        )

    @property
    def _str_extra(self):
        return super()._str_extra.cat(self.commands)

    def shell(self, ident):
        pred = lambda a: isinstance(a, Shell) and a.has_ident(ident)
        return (self.commands.find(pred)
                .to_either(self.no_such_shell_error.format(ident)))

    def add_history(self, cmd):
        new = (
            self.history.cons(cmd)
            .distinct_by(_history_cmp)
            .take(self.max_history)
        )
        return self.set(history=new)

    def transient_command(self, cmd):
        return self.set(transient_cmd=Just(cmd))

    @property
    def latest_command(self):
        return self.history.head.to_either(self.no_latest_command_error)

__all__ = ('Commands', 'Command', 'VimCommand', 'ShellCommand', 'Shell',
           'TransientCommand')



from libtmux import formats
from libtmux.session import Session as LTSession
from libtmux.common import EnvironmentMixin

from ribosome.record import list_field, maybe_field, field

from amino import List, __, Maybe, Map
from amino.lazy import lazy

from myo.ui.tmux.adapter import Adapter
from myo.ui.tmux.window import WindowAdapter, Window, NativeWindow
from myo.ui.tmux.util import parse_session_id
from myo.record import Named
from myo.logging import Logging


class Session(Named):
    name = field(str)
    id = maybe_field(int)
    windows = list_field(Window)

    @property
    def desc(self):
        id = self.id / ' {}'.format | ''
        return 'S{}'.format(id)

    @property
    def _str_extra(self):
        return super()._str_extra + self.id.to_list + self.windows


class VimSession(Session):
    session_name = '<vim>'

    def __new__(cls, *a, name=None, **kw):
        n = Maybe(name) | VimSession.session_name
        return super().__new__(cls, *a, name=n, **kw)

    @property
    def desc(self):
        return 'V{}'.format(super().desc)


class NativeSession(LTSession, Logging):

    def __init__(self, server=None, **kwargs):
        EnvironmentMixin.__init__(self)
        self.server = server
        if 'session_id' not in kwargs:
            raise ValueError('Session requires a `session_id`')
        self._session_id = kwargs['session_id']
        self._data = Map(kwargs) + ('id', self._session_id)

    @property
    def _info(self):
        return self._data

    def _list_windows(self):
        wformats = ['session_name', 'session_id'] + formats.WINDOW_FORMATS
        tmux_formats = ['#{%s}' % format for format in wformats]
        proc = self.cmd('list-windows', '-F%s' % '\t'.join(tmux_formats))
        windows = proc.stdout
        wformats = ['session_name', 'session_id'] + formats.WINDOW_FORMATS
        windows = [dict(zip(
            wformats, window.split('\t'))) for window in windows]
        windows = [
            dict((k, v) for k, v in window.items() if v) for window in windows
        ]
        return windows

    def list_windows(self):
        return [NativeWindow(session=self, **window) for window in
                self._windows]


class SessionAdapter(Adapter):

    @lazy
    def id(self):
        return self.native.id

    @property
    def id_i(self):
        return parse_session_id(self.id)

    @lazy
    def windows(self):
        return List.wrap(self.native.windows) / WindowAdapter

    def window_by_id(self, id):
        return self.windows.find(__.id_i.contains(id))

    def find_pane_by_id(self, id: int):
        return self.windows.find_map(__.pane_by_id(id))

    def new_window(self, **kw):
        return WindowAdapter(self.native.new_window(**kw))

    @property
    def attached(self):
        return int(self.native['session_attached']) > 0

__all__ = ('Session',)



import re

from amino import Maybe, List, __, _
from amino.test import later
from amino.test.path import fixture_path

from ribosome.record import encode_json

from myo.output.data import OutputEntry, OutputEvent, ParseResult
from myo.command import TransientCommand, ShellCommand

from integration._support.tmux import TmuxCmdSpec
from integration._support.vimtest import vimtest

_parse_head = 'parsed output:'
_event_head = 'event 1:'


def _parse_echo(data):
    matcher = re.compile('^line (\d+)')
    matches = data / matcher.match // Maybe / __.group(1)
    entries = matches / (lambda a: OutputEntry(text='entry {}'.format(a)))
    event = OutputEvent(head=List(_event_head), entries=entries)
    return ParseResult(head=List(_parse_head), events=List(event))


class TmuxParseSpec(TmuxCmdSpec):

    def custom(self):
        l3 = 'line 3'
        lines = List('line 1', 'line 2', l3)
        s = lines.mk_string('\\n')
        heads = List(_parse_head, _event_head)
        target = heads + lines / __.replace('line', 'entry')
        self._create_command(
            'com',
            "echo '{}'".format(s),
            target='make',
            parser='py:integration.tmux.parse_spec._parse_echo'
        )
        self._open_pane('make')
        self.json_cmd('MyoRun com')
        self._output_contains(l3)
        self.json_cmd('MyoParse')
        check = lambda: self.vim.buffer.content.should.equal(target)
        later(check)

    def shell(self):
        self.vim.vars.set_p('jump_to_error', False)
        self.vim.vars.set_p('lang_reifier', False)
        err = List.random_string()
        line = 'raise Exception(\'{}\')'.format(err)
        target = 'Exception: {}'.format(err)
        self._py_shell()
        self._create_command('test', line, shell='py', langs=['python'])
        self.json_cmd_sync('MyoRun test')
        self._output_contains(target)
        self.json_cmd_sync('MyoRun test')
        later(lambda: self._output.filter(_ == target).should.have.length_of(2)
              )
        self.json_cmd_sync('MyoParse')
        text = '  File "<stdin>", line 1, in <module>'
        later(lambda: self.vim.buffer.content.head.should.contain(text))


class PythonVimTestBase(TmuxCmdSpec):

    @property
    def _fname(self):
        return fixture_path('tmux', 'vim_test', 'test.py')

    @property
    def _target(self):
        return 'RuntimeError: supernova'

    def _set_vars(self):
        super()._set_vars()
        self.vim.vars.set_p('tmux_no_watcher', True)
        self.vim.vars.set_p('jump_to_error', False)
        self.vim.vars.set('test#python#runner', 'nose')
        self.vim.vars.set_p('output_reifier',
                            'py:myo.output.reifier.base.LiteralReifier')

    def _pre_start(self):
        super()._pre_start()
        self.vim.cmd_sync('noswapfile edit {}'.format(self._fname))
        self.vim.buffer.vars.set_p('test_langs', ['python'])

    def _run_test(self):
        self.vim.cursor(4, 1)
        len_pre = self._output.length
        self.json_cmd('MyoVimTest')
        self._check_test(len_pre)

    def _check_test(self, len_pre):
        later(lambda: self._output.length.should.be.greater_than(len_pre))
        later(lambda: self._output.should.contain(self._target))

    def _run_parse(self):
        check = lambda: self.vim.buffer.content.last.should.contain(
            self._target)
        self.json_cmd_sync('MyoParse')
        later(check)


class PythonVimTestSpec(PythonVimTestBase):

    @vimtest
    def complete(self):
        self._run_test()
        self._run_parse()
        self.vim.cursor(6, 1)
        self.vim.cmd_sync('call feedkeys("\\<cr>")')
        check1 = lambda: self.vim.buffer.name.should.equal(str(self._fname))
        later(check1)
        self.vim.focus(1).run_sync()
        self.vim.cmd_sync('call feedkeys("q")')
        check2 = lambda: self.vim.buffers.should.have.length_of(1)
        later(check2)

    @vimtest
    def twice(self):
        self._run_test()
        self._run_parse()
        self.vim.window.close()
        self._run_test()
        self._run_parse()


class PythonVimTestLoadHistorySpec(PythonVimTestBase):

    @property
    def _line(self):
        return 'nosetests {}:Namespace.test_something'.format(self._fname)

    def _set_vars(self):
        super()._set_vars()
        cmd = TransientCommand(line=self._line,
                               command=ShellCommand(line='', name='<test>'),
                               transient=True,
                               name='test_line_<test>_VIroI',
                               langs=List('python'))
        history = encode_json([cmd]).get_or_raise
        self.vim.vars.set('Myo_history', history)

    @vimtest
    def history(self):
        len_pre = self._output.length
        self.json_cmd('MyoRunLatest')
        self._check_test(len_pre)
        self._run_parse()

__all__ = ('TmuxParseSpec', 'PythonVimTestSpec')



from ribosome.machine import may_handle, handle
from ribosome.unite import UniteKind
from ribosome.machine.base import UnitTask
from ribosome.unite.data import UniteSyntax
from ribosome.unite.machine import UniteMachine, UniteTransitions

from amino import Map, List, _, __
from amino.lazy import lazy

from myo.plugins.unite.message import UniteHistory
from myo.state import MyoTransitions, MyoComponent
from myo.plugins.unite.data import UniteNames, HistorySource


class MyoUniteTransitions(UniteTransitions, MyoTransitions):

    def unite_cmd(self, cmd):
        args = ' '.join(self.msg.unite_args)
        self.vim.cmd('Unite {} {}'.format(cmd, args))

    @may_handle(UniteHistory)
    def history(self):
        self.unite_cmd(UniteNames.history)


class Plugin(UniteMachine, MyoComponent):
    Transitions = MyoUniteTransitions

    def __init__(self, *a, **kw):
        UniteMachine.__init__(self)
        MyoComponent.__init__(self, *a, **kw)

    @lazy
    def history_source(self):
        return HistorySource()

    @property
    def sources(self):
        return List(self.history_source)

    @lazy
    def run_action(self):
        return Map(name='run', handler=UniteNames.run, desc='run command')

    @property
    def actions(self):
        return List(self.run_action)

    @lazy
    def run_kind(self):
        return UniteKind(UniteNames.history, List(self.run_action))

    @property
    def kinds(self):
        return List(self.run_kind)

__all__ = ('Plugin',)



from typing import Tuple

from networkx import DiGraph

from amino.lazy import lazy
from amino import List, Empty, Just, Maybe, __, L, _

from ribosome.record import field, maybe_field

from myo.output.data import (PositionEntry, ErrorEntry, OutputEntry,
                             OutputEvent, OutputLine, MultiEvent, CodeEntry)
from myo.output.parser.base import EdgeData, SimpleParser


class FileEntry(PositionEntry):
    fun = maybe_field(str)
    code = maybe_field(CodeEntry)

    def lines(self, event: OutputEvent, group=Empty()):
        x = self.code / L(OutputLine.create)(_.text, self)
        return super().lines(event) + x.to_list

    @staticmethod
    def create(text, path, line, fun=None):
        return FileEntry(text=text, path=path, line=line, fun=Maybe(fun))


class PyErrorEntry(ErrorEntry):
    exc = field(str)


class ColEntry(OutputEntry):
    ws = field(str)

    def lines(self, event: OutputEvent, group=Empty()):
        return List()

_file = EdgeData(
    r='  File "(?P<path>.+)", line (?P<line>\d+)(?:, in (?P<fun>\S+))?',
    entry=FileEntry.create
)
_code = EdgeData(r='^    (?P<code>.+)', entry=CodeEntry)
_error = EdgeData(r='^(?P<exc>\S+): (?P<error>.+)', entry=PyErrorEntry)
_col = EdgeData(r='^    (?P<ws>\s+)\^', entry=ColEntry)


class Parser(SimpleParser):

    @lazy
    def graph(self):
        g = DiGraph()
        g.add_edge('start', 'file', data=_file)
        g.add_edge('file', 'code', data=_code)
        g.add_edge('code', 'file', data=_file)
        g.add_edge('code', 'error', data=_error)
        g.add_edge('code', 'col', data=_col)
        g.add_edge('col', 'error', data=_error)
        return g

    def event(self, entries: List[OutputEntry]):
        def folder(z: Tuple[List[OutputEntry], Maybe[FileEntry]], a):
            res, cur = z
            add, new = (
                (cur, Just(a))
                if isinstance(a, FileEntry) else
                ((cur / __.set(code=Just(a))).or_else(Just(a)), Empty())
                if isinstance(a, CodeEntry) else
                (Just(a), Empty())
            )
            return res + add.to_list, new
        grouped, rest = entries.fold_left((List(), Empty()))(folder)
        complete = grouped + rest.to_list
        return Just(MultiEvent(entries=complete))

__all__ = ('Parser',)



import time
import signal
from typing import Tuple

from psutil import Process

from amino import (L, Task, _, Maybe, F, __, Either, I, Right, Left, List,
                   Empty, Try)
from amino.lazy import lazy
from amino.task import task

from ribosome.machine.transition import Fatal, NothingToDo
from ribosome.machine.state import Info

from myo.logging import Logging
from myo.ui.tmux.session import Session
from myo.ui.tmux.window import Window, WindowAdapter
from myo.ui.tmux.pane import Pane, PaneAdapter
from myo.ui.tmux.view_path import ViewPath
from myo.ui.tmux.layout import Layout
from myo.command import ShellCommand
from myo.ui.tmux.view import View


pane_not_found_error = 'pane not found: {}'


class WindowFacade(Logging):

    def __init__(self, session: Session, window: Window, adapter:
                 WindowAdapter) -> None:
        self.session = session
        self.window = window
        self.adapter = adapter

    @property
    def window_size(self):
        return self.adapter.size

    @lazy
    def panes(self):
        return self.adapter.panes

    @property
    def root(self):
        return self.window.root

    def ref_pane_fatal(self, layout: Layout) -> Task[Pane]:
        ''' A pane in the layout to be used to open a pane.
        Uses the first open pane available.
        Failure if no panes are open.
        '''
        return (Task.call(self.ref_pane, layout) //
                L(Task.from_maybe)(_, "no ref pane for {}".format(layout)))

    def ref_pane(self, layout: Layout) -> Maybe[Pane]:
        return (
            self._opened_panes(layout.panes).sort_by(_.position | 0).last
            .o(layout.layouts.find_map(self.ref_pane))
        )

    def native_ref_pane(self, layout: Layout):
        return self.ref_pane(layout) // self.native_pane

    def native_ref_pane_task(self, layout: Layout):
        return self.ref_pane_fatal(layout) // self.native_pane_task_fatal

    def native_pane(self, pane: Pane):
        return ((pane.id // self.adapter.pane_by_id)
                .to_either(pane_not_found_error.format(pane.name)))

    def native_pane_task(self, pane: Pane):
        return Task.call(self.native_pane, pane)

    def native_pane_task_fatal(self, pane: Pane):
        return self.native_pane_task(pane).join_either

    def open_pane(self, path: ViewPath) -> Task[Either[str, ViewPath]]:
        v = path.view
        def open(pane):
            def apply(layout):
                set = path.is_pane.c(lambda: I, lambda: v.replace_pane)
                return (
                    self._split_pane(layout, pane) /
                    set /
                    path.setter.view /
                    Right
                )
            return (self._pane_opener(path, pane).eff(Either) // apply).value
        view = (Task.now(v) if path.is_pane else
                v.panes.head.task('{} has no panes'.format(v)))
        return view // open

    def _pane_opener(self, path: ViewPath, pane: Pane
                     ) -> Task[Either[str, Layout]]:
        def find_layout():
            return (path.layouts.reversed
                    .find(self.layout_open)
                    .to_either(
                        Fatal('no open layout when trying to open pane')))
        return Task.now(
            self.view_open(pane).no.flat_either_call(
                Left(NothingToDo('{!r} already open'.format(pane))),
                find_layout,
            )
        )

    def _split_pane(self, layout, pane) -> Task[Pane]:
        return (
            self.native_ref_pane_task(layout) /
            __.split(layout.horizontal) /
            PaneAdapter /
            pane.open
        )

    def pack_path(self, path: ViewPath):
        return self.pack / __.replace(path)

    def pack(self):
        from myo.ui.tmux.pack import WindowPacker
        return WindowPacker(self).run

    def run_command_line(self, pane: Pane, line: str):
        return (
            self.native_pane_task_fatal(pane) /
            __.run_command(line)
        )

    def ensure_not_running(self, pane: Pane, kill=False, signals=List('kill')):
        def handle(pa):
            return (
                Task.now(Right(True))
                if pa.not_running else
                self._kill_process(pa, signals=signals)
                if pane.kill or kill else
                Task.now(Left('command already running'))
            )
        return self.native_pane_task_fatal(pane) // handle

    def command_pid(self, pane: Pane):
        return self.native_pane_task_fatal(pane) / _.command_pid

    def kill_process(self, pane, signals):
        return (
            self.native_pane_task(pane)
            .eff(Either) //
            L(self._kill_process)(_, signals)
        ).value

    @task
    def _kill_process(self, adapter, signals):
        def _wait_killed(timeout):
            start = time.time()
            while time.time() - start > timeout and adapter.running:
                time.sleep(.1)
        def kill(signame):
            if adapter.running:
                sig = getattr(signal, 'SIG{}'.format(signame.upper()),
                              signal.SIGINT)
                (
                    adapter.command_pid //
                    L(Try)(Process, _) //
                    L(Try)(__.send_signal, sig)
                )
                _wait_killed(3)
            return adapter.not_running
        def final_check():
            return adapter.running.e(
                'could not kill running process',
                Info('no process running')
            )
        return (
            signals.find(kill) /
            'process killed by signal {}'.format /
            Info /
            Right |
            final_check
        )

    def pipe(self, pane: Pane, base: str):
        lg = lambda p: self.log.debug(
            'piping {} to {}'.format(p.name, p.log_path | 'nowhere'))
        return (
            self.native_pane_task_fatal(pane) /
            __.pipe(base) /
            pane.setter.log_path %
            lg
        )

    @property
    def pane_ids(self):
        return self.panes // _.id_i

    def pane_open(self, pane: Pane):
        return pane.id.exists(self.pane_ids.contains)

    def layout_open(self, layout: Layout):
        return (layout.panes.exists(self.pane_open) or
                layout.layouts.exists(self.layout_open))

    def view_open(self, view: View):
        f = self.layout_open if isinstance(view, Layout) else self.pane_open
        return f(view)

    is_open = view_open

    def open_views(self, layout: Layout):
        return layout.views.filter(self.view_open)

    def open_pane_path(self, path: ViewPath) -> Task[Either[str, ViewPath]]:
        ''' legacy version
        '''
        p = path.view
        return (Task.now(Left('pane {} already open'.format(p)))
                if self.is_open(p)
                else self._open_pane_path(path))

    def _open_in_layouts_path(self, pane, layout, outer):
        if self.layout_open(layout):
            return self._open_in_layout(pane, layout) / (_ + (outer,))
        else:
            return (
                outer.detach_last.map2(F(self._open_in_layouts_path, pane)) |
                Task.failed('cannot open {} without open layout'.format(pane))
            ).map3(lambda p, l, o: (p, layout, o.cat(l)))

    def _open_in_layout(self, pane, layout) -> Task[Tuple[Layout, Pane]]:
        return self._split_pane(layout, pane) & (Task.now(layout))

    def close_pane_path(self, path: ViewPath):
        new_path = path.map_view(__.set(id=Empty()))
        return self.close(path.view) / __.lmap(Fatal).replace(new_path)

    def _opened_panes(self, panes):
        return panes.filter(self.is_open)

    def run_command(self, pane: Pane, command: ShellCommand):
        return self.run_command_line(pane, command.line)

    def close(self, pane: Pane):
        return (self.native_pane_task(pane).eff(Either) //
                __.kill.map(Right)).value

    def focus(self, pane: Pane):
        return (
            self.native_pane(pane)
            .task('pane not found for focus: {}'.format(pane)) /
            __.focus()
        )

__all__ = ('WindowFacade',)



from ribosome.machine import message, json_message

RunVimCommand = message('RunVimCommand', 'command')
AddDispatcher = message('AddDispatcher')
StageI = message('StageI')
StageII = message('StageII')
Initialized = message('Initialized')
Parse = json_message('Parse')
ParseOutput = json_message('ParseOutput', 'command', 'output', 'path')
Resized = message('Resized')

__all__ = ('RunVimCommand', 'AddDispatcher', 'StageI', 'StageII',
           'Initialized', 'ParseOutput', 'Parse', 'Resized')



from amino import Task, L, Just, _

from ribosome.unite import UniteSource


class UniteNames():
    history_candidates = '_myo_unite_history'
    run = '_myo_unite_run_project'
    history = 'myo_history'
    syntax = '_myo_unite_syntax'


class HistorySource(UniteSource):

    def __init__(self) -> None:
        super().__init__(UniteNames.history, UniteNames.history_candidates,
                         UniteNames.history, Just(UniteNames.syntax))

    def syntax_task(self, syntax):
        m = L(Task.call)(syntax.match, _, _)
        hi = L(Task.call)(syntax.highlight, _)
        name = '^\s*\[\S\+\]'
        return (
            m('name', name) +
            hi('name', ctermfg=3) +
            m('line', '\({} \)\@<=.*'.format(name)) +
            hi('line', ctermfg=4)
        )


__all__ = ('HistorySource', 'UniteNames')



from amino import List


class View():
    pass


class Views():

    def __init__(self, views: List[View]=List()) -> None:
        self.views = views

    def __add__(self, view: View):
        return Views(self.views + [view])

__all__ = ('Views', 'View')
